/*
 * Basic hashmap implementation.
 *
 * Key-value pairs can be inserted into and removed from the hashmap.
 * Values may be of type var (fixed-point numbers, pointers and handles).
 * Character strings are used as keys.
 */

#include "hashmap.h"

// Gets the hash for the given key.
#define HASHMAP_HASH(_key_) hashmap_hash(_key_) % (self)->capacity

// OPrototypes of static functions.
HASHMAP_NODE *hashmap_create_node(var value, char *key);
HASHMAP_NODE *hashmap_insert_node(HASHMAP *self, HASHMAP_NODE *new_node);
HASHMAP_NODE *hashmap_get_node(HASHMAP *self, char *key);
void hashmap_rehash(HASHMAP *self);
int hashmap_hash(char *key);


// Creates a new hashmap.
HASHMAP *hashmap_create(int capacity)
{
    HASHMAP *self = sys_malloc(sizeof(HASHMAP));

    if (self)
    {
        self->capacity = capacity;
        if (self->capacity < HASHMAP_MIN_CAPACITY)
            self->capacity = HASHMAP_MIN_CAPACITY;
        self->nodes = sys_malloc(self->capacity * sizeof(HASHMAP_NODE*));
        
        if (!self->nodes)
        {
            sys_free(self);
            return NULL;
        }
    }

    return self;
}

HASHMAP *hashmap_create()
{
    return hashmap_create(HASHMAP_MIN_CAPACITY);
}

// Removes the hashmap.
void hashmap_free(HASHMAP *self)
{
    hashmap_remove_all(self);
    sys_free(self->nodes);
    sys_free(self);
}

// Inserts a key-value pair into the hashmap.
HASHMAP_NODE *hashmap_insert(HASHMAP *self, var value, char *key)
{
    if (hashmap_contains(self, key))
        return NULL;
    
    float load_factor = (float)self->length / (float)self->capacity;
    
    if (load_factor >= HASHMAP_LOAD_FACTOR)
        hashmap_rehash(self);
    
    return hashmap_insert_node(self, hashmap_create_node(value, key));
}

// Checks if the hashmap contains a certain key.
BOOL hashmap_contains(HASHMAP *self, char *key)
{
    return (hashmap_get_node(self, key) != NULL);
}

// Gets the value that is associated with the given key.
var hashmap_get(HASHMAP *self, char *key)
{
    HASHMAP_NODE *node = hashmap_get_node(self, key);
    if (node)
        return node->value;
    return 0;
}

// Removes the value that is associated with the given key.
// The removed value is returned.
var hashmap_remove(HASHMAP *self, char *key)
{
    HASHMAP_NODE *node = hashmap_get_node(self, key);
    
    if (!node) return 0;
    
    if (node->prev)
        node->prev->next = node->next;
    else
        (self->nodes)[HASHMAP_HASH(key)] = node->next;
    
    if (node->next)
        node->next->prev = node->prev;

    self->length--;
    
    var value = node->value;
    sys_free(node);
    return value;
}

// Removes all key-value pairs from the hashmap.
void hashmap_remove_all(HASHMAP *self)
{
    int i;
    for (i = 0; i < self->capacity; i++)
    {
        HASHMAP_NODE *node = (self->nodes)[i];
        
        while (node)
        {
            HASHMAP_NODE* next_node = node->next;
            sys_free(node);
            node = next_node;
        }
        
        (self->nodes)[i] = NULL;
    }
    
    self->length = 0;
}

// For iterating over all nodes in random order.
// Initial value of "index" should be 0. Initial value of "node" should be NULL.
// Returns the next node or NULL when all nodes have been iterated.
HASHMAP_NODE *hashmap_iterate(HASHMAP *self, int *index, HASHMAP_NODE **node)
{
    if (*node != NULL)
    {
        *node = (*node)->next;
        if (!*node) *index++; 
    }

    while (!*node && *index < self->capacity)
    {
        if (*node = (self->nodes)[*index])
            return *node;
        *index++;
    }

    return *node;
}

// Used internally for creating new nodes.
HASHMAP_NODE *hashmap_create_node(var value, char *key)
{
    HASHMAP_NODE *node = sys_malloc(sizeof(HASHMAP_NODE));
    
    if (node)
    {
        node->value = value;
        strcpy(node->key, key);
    }
    
    return node;
}

// Used internally for inserting nodes into the hashmap.
HASHMAP_NODE *hashmap_insert_node(HASHMAP *self, HASHMAP_NODE *new_node)
{
    if (!new_node) return NULL;

    int hash = HASHMAP_HASH(new_node->key);
    HASHMAP_NODE *node = (self->nodes)[hash];

    if (!node)
    {
        (self->nodes)[hash] = new_node;
        new_node->prev = new_node->next = NULL;
    }
    else
    {
        while (node->next)
            node = node->next;
        node->next = new_node;
        new_node->prev = node;
        new_node->next = NULL;
    }
    
    self->length++;
    return new_node;
}

// Used internally for getting the node that is associated with a certain key.
HASHMAP_NODE *hashmap_get_node(HASHMAP *self, char *key)
{
    HASHMAP_NODE *node = (self->nodes)[HASHMAP_HASH(key)];

    while (node)
    {
        if (strcmp(node->key, key) == 0)
            break;
        node = node->next;
    }
    
    return node;
}

// Used internally for resizing the hashmap.
void hashmap_rehash(HASHMAP *self)
{
    HASHMAP *temp_hashmap = hashmap_create(self->capacity * 2);
    
    // insert nodes
    int i;
    for (i = 0; i < self->capacity; i++)
    {
        HASHMAP_NODE *node = (self->nodes)[i];
        while (node)
        {
            HASHMAP_NODE* next_node = node->next;
            hashmap_insert_node(temp_hashmap, node);
            node = next_node;
        }
    }
    
    sys_free(self->nodes);
    memcpy(self, temp_hashmap, sizeof(HASHMAP));
    sys_free(temp_hashmap);
}

// SDBM hash function.
int hashmap_hash(char *key)
{
    int hash = 0;
    for ( ; *key; key++)
        hash = *key + (hash << 6) + (hash << 16) - hash;
    return abs(hash);
}