/*
 * Basic hashmap implementation.
 *
 * Key-value pairs can be inserted into and removed from the hashmap.
 * Values may be of type var (fixed-point numbers, pointers and handles).
 * Character strings are used as keys.
 */

#ifndef HASHMAP_H
#define HASHMAP_H

// You can't create smaller hashmaps than this.
#define HASHMAP_MIN_CAPACITY 16

// How many characters each key may have (including '\0').
#define HASHMAP_KEY_LENGTH 32

#define HASHMAP_LOAD_FACTOR 0.75

typedef struct HASHMAP_NODE
{
    char key[HASHMAP_KEY_LENGTH];
    var value;
    struct HASHMAP_NODE *prev, *next;
} HASHMAP_NODE;

typedef struct
{
    int length; // number of nodes
    int capacity;
    HASHMAP_NODE **nodes; // pointer array
} HASHMAP;


// Creates a new hashmap.
HASHMAP *hashmap_create(int capacity);
HASHMAP *hashmap_create();

// Removes the hashmap.
void hashmap_free(HASHMAP *self);

// Inserts a key-value pair into the hashmap.
HASHMAP_NODE *hashmap_insert(HASHMAP *self, var value, char *key);

// Checks if the hashmap contains a certain key.
BOOL hashmap_contains(HASHMAP *self, char *key);

// Gets the value that is associated with the given key.
var hashmap_get(HASHMAP *self, char *key);

// Removes the value that is associated with the given key.
// The removed value is returned.
var hashmap_remove(HASHMAP *self, char *key);

// Removes all key-value pairs from the hashmap.
void hashmap_remove_all(HASHMAP *self);

// For iterating over all nodes in random order.
// Initial value of "index" should be 0. Initial value of "node" should be NULL.
// Returns the next node or NULL when all nodes have been iterated.
HASHMAP_NODE *hashmap_iterate(HASHMAP *self, int *index, HASHMAP_NODE **node);

#endif // HASHMAP_H
